/*****************************************************************************
 * file:        $RCSfile: fakeidxscan.c,v $ $Revision: 1.8 $
 * module:      fakeidxscan
 * authors:     jeremyd
 * last mod:    $Author: jeremyd $ at $Date: 2006/01/19 22:05:26 $
 *
 * created:     Mon Dec 12 00:27:42 PST 2005
 *
 *****************************************************************************/

#include <string.h>
#include <math.h>

#include "postgres.h"
#include "access/genam.h"
#include "access/heapam.h"
#include "access/nbtree.h"
#include "access/tuptoaster.h"
#include "catalog/catalog.h"
#include "catalog/indexing.h"
#include "catalog/pg_largeobject.h"
#include "commands/comment.h"
#include "utils/lsyscache.h"
#include "utils/typcache.h"
#include "utils/resowner.h"


#include "fmgr.h"
#include "libpq/pqformat.h"
#include "utils/builtins.h"
#include "funcapi.h"

Datum fakeidxrowstst(PG_FUNCTION_ARGS);
Datum fakeidxrowscomposite(PG_FUNCTION_ARGS);
Datum fakeidxrowsset(PG_FUNCTION_ARGS);

typedef struct fakeidxscan_state
{
	Relation rel;
	Relation idx;
	IndexScanDesc scan;
	int32 num_ret_cols;
	int16 * ret_col_nums;
	IndexTuple itup;
	TupleDesc itupdesc;
	int32 ret_col_iter;
	ScanKeyData keys[1];
} fakeidxscan_state;

#define FAKEIDXSCAN_STATE_SIZE(n) (sizeof(fakeidxscan_state) + (sizeof(ScanKeyData) * (n - 1)))

/*
 * Going to need
 *
 * relaltionid
 * indexid
 * col1num
 * col2num
 * col1val
 * col2min
 * col2max
 */

#define DIMINVAL(x) ((x)->ndim != 1)
#define SIZEINVAL(x) ((x)->dim1 != numscankeys)

/* callback for ExprContext - promises to call me on func scan shutdown */
static void fakeidxrowstst_cleanup_cb (Datum arg)
{
	fakeidxscan_state * state = (fakeidxscan_state *) DatumGetPointer(arg);

	index_endscan(state->scan);
	index_close(state->idx);
	heap_close(state->rel, AccessShareLock);
	pfree(state->ret_col_nums);
	pfree(state);
}

/** TODO: error handling - bad params == bad server crash */
/* some pseudocode assumptions about parameters:
 * length(colnums) == length(opers) == length(strategies) == nattrs in vals == numscankeys
 * 0 < numscankeys <= nattrs in index
 * 0 < retcolnum <= nattrs in index
 * colnums must begin at 1 and be in order with no gaps (repeats ok)
 * better not try anything funky with the opers vs strategies - not sure what will happen there
 * probably more
 */
PG_FUNCTION_INFO_V1(fakeidxrowstst);
Datum
fakeidxrowstst(PG_FUNCTION_ARGS)
{
	FuncCallContext *funcctx;
	MemoryContext    oldcontext;
	fakeidxscan_state * state;


	if (SRF_IS_FIRSTCALL())
	{
		Oid myrelid = PG_GETARG_OID(0);
		Oid myidxid = PG_GETARG_OID(1);
		//int32 numcounted = 0;
		int32 numscankeys = PG_GETARG_INT32(2);
		int2vector * colnums = (int2vector *) PG_GETARG_POINTER(4);
		oidvector * opers = (oidvector *) PG_GETARG_POINTER(5);
		int2vector * strategies = (int2vector *) PG_GETARG_POINTER(6);
		HeapTupleHeader vals = PG_GETARG_HEAPTUPLEHEADER(7);

		/* crap from GetAttributeByNum, since it appears to be sucky to
		 * call it in a loop */
		Oid			tupType;
		int32		tupTypmod;
		TupleDesc	tupDesc;
		HeapTupleData tmptup;
		int i;
		bool isNull;

		/* validate some parameters */
		if (DIMINVAL(colnums) || DIMINVAL(opers) || DIMINVAL(strategies))
		{
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_COLUMNS),
					 errmsg("array params must be one dimensional"),
					 errhint("fix it")));
		}

		if (SIZEINVAL(colnums) || SIZEINVAL(opers) || SIZEINVAL(strategies) || (numscankeys == 1) ? (vals->t_natts < 1) : (vals->t_natts != (int16)numscankeys))
		{
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
					 errmsg("array size must be the same as the specified numkeys"),
					 errhint("fix it")));
		}

		tupType = HeapTupleHeaderGetTypeId(vals);
		tupTypmod = HeapTupleHeaderGetTypMod(vals);
		tupDesc = lookup_rowtype_tupdesc(tupType, tupTypmod);
		/*
		 * heap_getattr needs a HeapTuple not a bare HeapTupleHeader.  We set all
		 * the fields in the struct just in case user tries to inspect system
		 * columns.
		 */
		tmptup.t_len = HeapTupleHeaderGetDatumLength(vals);
		ItemPointerSetInvalid(&(tmptup.t_self));
		tmptup.t_tableOid = InvalidOid;
		tmptup.t_data = vals;

		funcctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		state = (fakeidxscan_state *)palloc (FAKEIDXSCAN_STATE_SIZE(numscankeys));
		if (!state)
		{
			ereport(ERROR,
					(errcode(ERRCODE_OUT_OF_MEMORY),
					 errmsg("unable to allocate multicall state struct"),
					 errhint("you suck")));
			PG_RETURN_NULL();
		}

		state->rel = heap_open (myrelid, AccessShareLock);
		state->idx = index_open (myidxid);
		state->num_ret_cols = 1;
		state->ret_col_nums = (int16 *) palloc (sizeof(int16));
		state->ret_col_nums[0] = PG_GETARG_INT16(3);

		for (i=0; i<numscankeys; ++i)
		{
			Datum foo = heap_getattr(&tmptup, (AttrNumber) i+1, tupDesc, &isNull);
			if (!isNull)
			{
				ScanKeyInit(&state->keys[i], (AttrNumber) colnums->values[i], (StrategyNumber) strategies->values[i], get_opcode(opers->values[i]), foo);
			}
		}

		state->scan = index_beginscan(state->rel, state->idx, SnapshotNow, numscankeys, state->keys);

		funcctx->user_fctx = (void *)state;

		RegisterExprContextCallback(((ReturnSetInfo *) fcinfo->resultinfo)->econtext,
				&fakeidxrowstst_cleanup_cb, PointerGetDatum(state));

		MemoryContextSwitchTo(oldcontext);
	}

	funcctx = SRF_PERCALL_SETUP();
	state = funcctx->user_fctx;

	if (index_getnext_indexitem(state->scan, ForwardScanDirection))
	{
		Datum data;
		/* this var may be hackish */
		TupleDesc tupdesc = state->idx->rd_att;
		/* this var is definately hackish */
		BTScanOpaque so = (BTScanOpaque)state->scan->opaque;
		/* deep black magic (from nbtsearch.c) */
		ItemPointer current = &(state->scan->currentItemData);
		Buffer buf = so->btso_curbuf;
		Page page = BufferGetPage(buf);
		IndexTuple itup = &(((BTItem) PageGetItem(page, PageGetItemId(page, ItemPointerGetOffsetNumber(current))))->bti_itup);
		bool isNull;

		/* this is hard-coded for the single element return */
		data = index_getattr(itup, (AttrNumber) state->ret_col_nums[0], tupdesc, &isNull);

		if (isNull)
		{
			/* not sure this will work. I can find no examples of
			 * SRFs returning null. */
			fcinfo->isnull = true;
			SRF_RETURN_NEXT (funcctx, (Datum) 0);
		}
		else
		{
			SRF_RETURN_NEXT (funcctx, data);
		}
	}

	SRF_RETURN_DONE (funcctx);
}

/** TODO: error handling - bad params == bad server crash */
/* some pseudocode assumptions about parameters:
 * length(colnums) == length(opers) == length(strategies) == nattrs in vals == numscankeys
 * 0 < numscankeys <= nattrs in index
 * 0 < retcolnum <= nattrs in index
 * colnums must begin at 1 and be in order with no gaps (repeats ok)
 * better not try anything funky with the opers vs strategies - not sure what will happen there
 * probably more
 */
PG_FUNCTION_INFO_V1(fakeidxrowscomposite);
Datum
fakeidxrowscomposite(PG_FUNCTION_ARGS)
{
	FuncCallContext *funcctx;
	MemoryContext    oldcontext;
	fakeidxscan_state * state;


	if (SRF_IS_FIRSTCALL())
	{
		Oid myrelid = PG_GETARG_OID(0);
		Oid myidxid = PG_GETARG_OID(1);
		//int32 numcounted = 0;
		int32 numscankeys = PG_GETARG_INT32(2);
		int2vector * retcolnums = (int2vector *) PG_GETARG_POINTER(3);
		int2vector * colnums = (int2vector *) PG_GETARG_POINTER(4);
		oidvector * opers = (oidvector *) PG_GETARG_POINTER(5);
		int2vector * strategies = (int2vector *) PG_GETARG_POINTER(6);
		HeapTupleHeader vals = PG_GETARG_HEAPTUPLEHEADER(7);
		TupleDesc	rettupdesc;

		/* crap from GetAttributeByNum, since it appears to be sucky to
		 * call it in a loop */
		Oid			tupType;
		int32		tupTypmod;
		TupleDesc	tupDesc;
		HeapTupleData tmptup;
		int i;
		bool isNull;

		/* validate some parameters */
		if (DIMINVAL(colnums) || DIMINVAL(opers) || DIMINVAL(strategies) || DIMINVAL(retcolnums))
		{
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_COLUMNS),
					 errmsg("array params must be one dimensional"),
					 errhint("fix it")));
		}

		if (SIZEINVAL(colnums) || SIZEINVAL(opers) || SIZEINVAL(strategies) || (numscankeys == 1) ? (vals->t_natts < 1) : (vals->t_natts != (int16)numscankeys))
		{
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
					 errmsg("array size must be the same as the specified numkeys"),
					 errhint("fix it")));
		}

		if (get_call_result_type(fcinfo, NULL, &rettupdesc) != TYPEFUNC_COMPOSITE)
			ereport(ERROR,
					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
					 errmsg("function returning record called in context "
						 "that cannot accept type record")));

		if (rettupdesc->natts != retcolnums->dim1)
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
					 errmsg("array size must be the same as the specified number of return columns"),
					 errhint("fix it")));

		tupType = HeapTupleHeaderGetTypeId(vals);
		tupTypmod = HeapTupleHeaderGetTypMod(vals);
		tupDesc = lookup_rowtype_tupdesc(tupType, tupTypmod);
		/*
		 * heap_getattr needs a HeapTuple not a bare HeapTupleHeader.  We set all
		 * the fields in the struct just in case user tries to inspect system
		 * columns.
		 */
		tmptup.t_len = HeapTupleHeaderGetDatumLength(vals);
		ItemPointerSetInvalid(&(tmptup.t_self));
		tmptup.t_tableOid = InvalidOid;
		tmptup.t_data = vals;

		funcctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		state = (fakeidxscan_state *)palloc (FAKEIDXSCAN_STATE_SIZE(numscankeys));
		if (!state)
		{
			ereport(ERROR,
					(errcode(ERRCODE_OUT_OF_MEMORY),
					 errmsg("unable to allocate multicall state struct"),
					 errhint("you suck")));
			PG_RETURN_NULL();
		}

		state->rel = heap_open (myrelid, AccessShareLock);
		state->idx = index_open (myidxid);
		state->num_ret_cols = retcolnums->dim1;
		state->ret_col_nums = (int16 *) palloc (state->num_ret_cols * sizeof(int16));

		memcpy(state->ret_col_nums, retcolnums->values, state->num_ret_cols * sizeof(int16));

		for (i=0; i<numscankeys; ++i)
		{
			Datum foo = heap_getattr(&tmptup, (AttrNumber) i+1, tupDesc, &isNull);
			if (!isNull)
			{
				ScanKeyInit(&state->keys[i], (AttrNumber) colnums->values[i], (StrategyNumber) strategies->values[i], get_opcode(opers->values[i]), foo);
			}
		}

		state->scan = index_beginscan(state->rel, state->idx, SnapshotNow, numscankeys, state->keys);

		funcctx->user_fctx = (void *)state;

		RegisterExprContextCallback(((ReturnSetInfo *) fcinfo->resultinfo)->econtext,
				&fakeidxrowstst_cleanup_cb, PointerGetDatum(state));

		funcctx->tuple_desc = BlessTupleDesc(rettupdesc);

		MemoryContextSwitchTo(oldcontext);
	}

	funcctx = SRF_PERCALL_SETUP();
	state = funcctx->user_fctx;

	if (index_getnext_indexitem(state->scan, ForwardScanDirection))
	{
		Datum * data = (Datum *) palloc(state->num_ret_cols * sizeof(Datum));
		bool * isNull = (bool *) palloc(state->num_ret_cols * sizeof(bool));
		HeapTuple rettuple;
		/* this var may be hackish */
		TupleDesc itupdesc = state->idx->rd_att;
		/* this var is definately hackish */
		BTScanOpaque so = (BTScanOpaque)state->scan->opaque;
		/* deep black magic (from nbtsearch.c) */
		ItemPointer current = &(state->scan->currentItemData);
		Buffer buf = so->btso_curbuf;
		Page page = BufferGetPage(buf);
		IndexTuple itup = &(((BTItem) PageGetItem(page, PageGetItemId(page, ItemPointerGetOffsetNumber(current))))->bti_itup);
		int i;

		for (i = 0; i < state->num_ret_cols; ++i)
		{
			data[i] = index_getattr(itup, (AttrNumber) state->ret_col_nums[i], itupdesc, &isNull[i]);
		}

		rettuple = heap_form_tuple(funcctx->tuple_desc, data, isNull);
		pfree(data);
		pfree(isNull);

		SRF_RETURN_NEXT(funcctx, HeapTupleGetDatum(rettuple));
	}
	else
	{
#if 0
		index_endscan(state->scan);
		index_close(state->idx);
		heap_close(state->rel, AccessShareLock);
#endif
	}

	SRF_RETURN_DONE (funcctx);
}


/*
 * This is pretty specific to one of our problems but may be useful in general.
 * Given the same parameters as the composite variant, returns each
 * column of each row that would have been returned in that one as
 * its own row.  so if you would have gotten 8 rows of 8 columns each
 * out of the composite function, you would get 64 rows out of this one.
 *
 * Also, this will swallow up the nulls for you.  That is, it will not return
 * any nulls.  If a row would have only had null columns, I think this may
 * break.
 *
 * XXX This rather depends on the (apparently currently correct) assumption
 * that tuples returned from the index scan will not be trashed with the
 * percall memory context, but will be maintained for the multicall context.
 * This appears to be a valid assumption because of various things I read in
 * the code about how the index pages are "pinned" while the scan is happening.
 */
PG_FUNCTION_INFO_V1(fakeidxrowsset);
Datum
fakeidxrowsset(PG_FUNCTION_ARGS)
{
	FuncCallContext *funcctx;
	MemoryContext    oldcontext;
	fakeidxscan_state * state;


	if (SRF_IS_FIRSTCALL())
	{
		Oid myrelid = PG_GETARG_OID(0);
		Oid myidxid = PG_GETARG_OID(1);
		//int32 numcounted = 0;
		int32 numscankeys = PG_GETARG_INT32(2);
		int2vector * retcolnums = (int2vector *) PG_GETARG_POINTER(3);
		int2vector * colnums = (int2vector *) PG_GETARG_POINTER(4);
		oidvector * opers = (oidvector *) PG_GETARG_POINTER(5);
		int2vector * strategies = (int2vector *) PG_GETARG_POINTER(6);
		HeapTupleHeader vals = PG_GETARG_HEAPTUPLEHEADER(7);

		/* crap from GetAttributeByNum, since it appears to be sucky to
		 * call it in a loop */
		Oid			tupType;
		int32		tupTypmod;
		TupleDesc	tupDesc;
		HeapTupleData tmptup;
		int i;
		bool isNull;

		/* validate some parameters */
		if (DIMINVAL(colnums) || DIMINVAL(opers) || DIMINVAL(strategies) || DIMINVAL(retcolnums))
		{
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_COLUMNS),
					 errmsg("array params must be one dimensional"),
					 errhint("fix it")));
		}

		if (SIZEINVAL(colnums) || SIZEINVAL(opers) || SIZEINVAL(strategies) || (numscankeys == 1) ? (vals->t_natts < 1) : (vals->t_natts != (int16)numscankeys))
		{
			ereport(ERROR,
					(errcode(ERRCODE_TOO_MANY_ARGUMENTS),
					 errmsg("array size must be the same as the specified numkeys"),
					 errhint("fix it")));
		}

		tupType = HeapTupleHeaderGetTypeId(vals);
		tupTypmod = HeapTupleHeaderGetTypMod(vals);
		tupDesc = lookup_rowtype_tupdesc(tupType, tupTypmod);
		/*
		 * heap_getattr needs a HeapTuple not a bare HeapTupleHeader.  We set all
		 * the fields in the struct just in case user tries to inspect system
		 * columns.
		 */
		tmptup.t_len = HeapTupleHeaderGetDatumLength(vals);
		ItemPointerSetInvalid(&(tmptup.t_self));
		tmptup.t_tableOid = InvalidOid;
		tmptup.t_data = vals;

		funcctx = SRF_FIRSTCALL_INIT();
		oldcontext = MemoryContextSwitchTo(funcctx->multi_call_memory_ctx);
		state = (fakeidxscan_state *)palloc (FAKEIDXSCAN_STATE_SIZE(numscankeys));
		if (!state)
		{
			ereport(ERROR,
					(errcode(ERRCODE_OUT_OF_MEMORY),
					 errmsg("unable to allocate multicall state struct"),
					 errhint("you suck")));
			PG_RETURN_NULL();
		}

		state->rel = heap_open (myrelid, AccessShareLock);
		state->idx = index_open (myidxid);
		state->num_ret_cols = retcolnums->dim1;
		state->ret_col_nums = (int16 *) palloc (state->num_ret_cols * sizeof(int16));
		state->ret_col_iter = state->num_ret_cols;

		memcpy(state->ret_col_nums, retcolnums->values, state->num_ret_cols * sizeof(int16));

		for (i=0; i<numscankeys; ++i)
		{
			Datum foo = heap_getattr(&tmptup, (AttrNumber) i+1, tupDesc, &isNull);
			if (!isNull)
			{
				ScanKeyInit(&state->keys[i], (AttrNumber) colnums->values[i], (StrategyNumber) strategies->values[i], get_opcode(opers->values[i]), foo);
			}
		}

		state->scan = index_beginscan(state->rel, state->idx, SnapshotNow, numscankeys, state->keys);

		funcctx->user_fctx = (void *)state;

		RegisterExprContextCallback(((ReturnSetInfo *) fcinfo->resultinfo)->econtext,
				&fakeidxrowstst_cleanup_cb, PointerGetDatum(state));

		MemoryContextSwitchTo(oldcontext);
	}

	funcctx = SRF_PERCALL_SETUP();
	state = funcctx->user_fctx;

	while (++(state->ret_col_iter) < state->num_ret_cols)
	{
		bool isNull;
		Datum data = index_getattr(state->itup, (AttrNumber) state->ret_col_nums[state->ret_col_iter], state->itupdesc, &isNull);
		if (isNull)
		{
			continue;
		}
		else
		{
			SRF_RETURN_NEXT(funcctx, data);
			break;
		}
	}

	if (index_getnext_indexitem(state->scan, ForwardScanDirection))
	{
		/* this var may be hackish */
		TupleDesc itupdesc = state->idx->rd_att;
		/* this var is definately hackish */
		BTScanOpaque so = (BTScanOpaque)state->scan->opaque;
		/* deep black magic (from nbtsearch.c) */
		ItemPointer current = &(state->scan->currentItemData);
		Buffer buf = so->btso_curbuf;
		Page page = BufferGetPage(buf);
		IndexTuple itup = &(((BTItem) PageGetItem(page, PageGetItemId(page, ItemPointerGetOffsetNumber(current))))->bti_itup);

		state->itupdesc = itupdesc;
		state->itup = itup;

		for (state->ret_col_iter = 0; state->ret_col_iter < state->num_ret_cols; ++state->ret_col_iter)
		{
			bool isNull;
			Datum data = index_getattr(state->itup, (AttrNumber) state->ret_col_nums[state->ret_col_iter], state->itupdesc, &isNull);
			if (!isNull)
				SRF_RETURN_NEXT(funcctx, data);
		}
	}
	else
	{
#if 0
		index_endscan(state->scan);
		index_close(state->idx);
		heap_close(state->rel, AccessShareLock);
#endif
	}

	SRF_RETURN_DONE (funcctx);
}
